import time
from melodically.harmony import midi_to_std


class MidiNoteQueue:
    """
    A midi queue containing note_on/off midi messages and timestamps
    (the data structure returned by the get_timestamp_msg function).
    This queue is used by the rhythmic parser algorithm.
    """

    def __init__(self):
        # container used to implement the queue
        self._container = []

        # timestamp of the last note_on message
        self._lastTimestamp = 0

        # list of midi values of the note_on messages that are not closed yet
        self._openNoteOnList = []

        # threshold in seconds to discard notes that are too close
        self._minimumInterval = 0.06

    def push(self, msg_type, note, timestamp=None):
        """
        Pushes a note_on/off message in the queue.
        If the note_on message is too close with the last note_on, the new entry is discarded.
        If a note_off message doesn't close a note_on message, the new entry is discarded.

        :param msg_type: 'note_on' or 'note_off'
        :param note: midi note value
        :param timestamp: optional timestamp value, if none is provided, it's calculated during the method execution
        """

        # getting the timestamp if none is provided
        if not timestamp:
            timestamp = time.time()

        # note_on case
        if msg_type == 'note_on':
            # checking if the pushed note_on is too close with the last one
            if timestamp - self._lastTimestamp > self._minimumInterval:
                self._lastTimestamp = timestamp
                self._openNoteOnList.append(note)
                self._container.append({
                    'type': msg_type,
                    'note': note,
                    'timestamp': timestamp
                })

        # note_off case
        elif msg_type == 'note_off':
            # checking if the note_off closes a note on
            if note in self._openNoteOnList:
                self._openNoteOnList.remove(note)
                self._container.append({
                    'type': msg_type,
                    'note': note,
                    'timestamp': timestamp
                })

        # all other types of midi messages are excluded automatically

    def pop(self):
        """
        Pops a midi message from the front of the queue.

        :return: midi message with timestamp
        """
        return self._container.pop(0)

    def get_container(self):
        """
        Getter for the container used for the queue.

        :return: list of midi messages with timestamp
        """
        return self._container

    def get_notes(self):
        """
        Gets a list of notes in standard notation from the note on messages.

        :return: list of notes in standard notation
        """
        notes = []
        self.clean_unclosed_note_ons()  # cleaning the container

        for msg in self._container:
            if msg['type'] == 'note_on':
                notes.append(midi_to_std(msg['note']))
        return notes

    def clean_unclosed_note_ons(self):
        """
        Removes from the queue the unclosed note_on messages.
        """
        for open_note in self._openNoteOnList:
            index = len(self._container) - 1  # the index the we're checking
            # searching for the most recent note_on with note == open_note
            while not (self._container[index]['type'] == 'note_on' and self._container[index]['note'] == open_note):
                index = index - 1
            del self._container[index]  # removing the open note_on message
        self._openNoteOnList.clear()  # removing the open note references

    def clear(self):
        """
        Removes all the elements from the queue.
        """
        self._container.clear()
        self._lastTimestamp = 0
        self._openNoteOnList = []



from melodically.rhythm import get_nearest_rhythm
from melodically.chords import chord_tones


def parse_musical_note(musical_note, chord):
    """
    Function that given a chord, parses a musical note
    into an abstract melody_parser notation.

    ============================
    MELODIC SYMBOLS
    ============================
    c: chord tone
    l: color tone
    x: random tone
    ============================

    :param musical_note: standard note notation
    :param chord: chord notation
    :return: abstract melody_parser note
    """
    if musical_note in chord_tones[chord]['c']:
        return 'c'
    elif musical_note in chord_tones[chord]['l']:
        return 'l'
    else:
        return 'x'


def parse_rhythm(midi_queue, rhythmical_durations):
    """
    Given a MidiNoteQueue object and a rhytmical_duration dictionary,
    parses the note messages contained into the MidiNoteQueue into
    a list of the following symbols, based on the durations of the various
    notes and rests.

    ============================
    RHYTHMIC SYMBOLS
    ============================
    1: whole note
    2: half note
    4: quarter note
    4dot: quarter note dotted
    4t: quarter note triplet
    8: eight note
    8t: eight note triplet
    16: sixteenth note
    16t: sixteenth note triplet

    the rests are indicated with an "r" suffix
    ex: r4 (quarter note rest)
    ============================

    :param midi_queue: MidiNoteQueue object
    :param rhythmical_durations: dictionary of harmonic durations
    :return: list of duration symbols
    """
    result = []  # this list will contain the rhythmical symbols
    midi_queue.clean_unclosed_note_ons()
    midi_queue_container = midi_queue.get_container()
    for i in range(len(midi_queue_container)):
        if midi_queue_container[i]['type'] == 'note_on':
            j = i + 1
            # finding the corresponding subsequent note_off
            while not (midi_queue_container[j]['type'] == 'note_off' and midi_queue_container[j]['note'] ==
                       midi_queue_container[i]['note']):
                j = j + 1
            # from here on j is the index of the note_off
            interval = midi_queue_container[j]['timestamp'] - midi_queue_container[i]['timestamp']

            # getting the closest rhythmic figure and adding it to the result
            rhythm = get_nearest_rhythm(interval, rhythmical_durations)
            result.append(rhythm)

            # computing also the rest
            if i == j - 1 and j < len(midi_queue_container) - 1:
                # no other note_ons between the current note_on and note_off
                rest_interval = midi_queue_container[j + 1]['timestamp'] - midi_queue_container[j]['timestamp']
                if rest_interval >= 0.08:
                    # rhythmic figure for the rest
                    rest_rhythm = get_nearest_rhythm(interval, rhythmical_durations)
                    result.append('r' + rest_rhythm)

    return result


def parse_melody(midi_queue, chord, rhythmical_durations):
    rhythmic_symbols = parse_rhythm(midi_queue, rhythmical_durations)
    notes = midi_queue.get_notes()
    note_count = 0
    result = []
    for i in range(len(rhythmic_symbols)):
        symbol = rhythmic_symbols[i]
        if 'r' not in symbol:
            # note detected
            symbol = parse_musical_note(notes[note_count], chord) + symbol
            note_count = note_count + 1
        result.append(symbol)

    return result



import numpy as np
from collections import Counter

"""
Musical notes.

This list defines, the musical notes in standard (std) notation.
The position of each notes in the list reflects the midi ordering.
"""
musical_notes = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']

"""
Musical notes with the flat notation
"""
musical_notes_b = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B']

"""
this data structure contains all the
possible semitone sequences to construct
many families of modal scales
"""
mode_signatures = [
    [2, 2, 1, 2, 2, 2, 1],  # TTSTTTS
    # [2, 2, 1, 3, 1, 2, 1],
]


def midi_to_std(midi_note):
    """
    Function that parses a midi note into a musical note in std notation.

    :param midi_note: midi note number
    :return: musical note
    """
    return musical_notes[midi_note % 12]


def std_to_midi(musical_note, octave=-1):
    """
    Convert a musical note into a midi note.

    :param musical_note: musical note symbol
    :param octave: octave of the note
    :return: midi note number
    """
    octave_offset = (octave + 1) * 12
    if musical_note in musical_notes:
        return musical_notes.index(musical_note) + octave_offset
    else:
        return musical_notes_b.index(musical_note) + octave_offset


def get_root(notes):
    """
    returns the most common value in the list of notes

    :param notes: notes in standard notation
    :return: single note in standard notation
    """
    return max(set(notes), key=notes.count)


def get_all_modes(root):
    """
    given a root, calculates all the modes

    :param root: note in std notation
    :return: multi dimensional list of size [len(mode_signatures)x7x7] containing the modes
    """
    modes_note_std = []
    for i in range(len(mode_signatures)):  # iterating for different modes families
        modes_note_std.append([])
        current_sequence = np.array(mode_signatures[i])
        for j in range(7):  # iterating for each mode in the family
            modes_note_std[i].append([root])
            last_note = musical_notes.index(root)
            for k in range(6):  # iterating for each note in the scale
                last_note = (last_note + current_sequence[k]) % 12
                modes_note_std[i][j].append(musical_notes[last_note])
            current_sequence = np.roll(current_sequence, -1)  # circular shift of the note sequence
    return modes_note_std


"""
dictionary containing all the notes as keys and
the respective modes as value
"""
modes_dict = {root: get_all_modes(root) for root in musical_notes}


def harmonic_affinities(root, notes_std):
    """
    Given a root and a set of notes, returns a multidimensional list of the
    size of the value of the modes_dict using the root as a key, that has as elements
    the result of a distance function applied to the various notes of each modal scale.

    :param root: root note in std notation
    :param notes_std: list of target notes in std notation
    :return: multidimensional list containing harmonic distances between notes_std and the modes of root
    """

    # affinity points used for each degree of a modal scale
    positive_weights = [1, 0.5, 3, 0.2, 0.4, 2, 0.1]

    # affinity points subtracted from the notes out of scale
    negative_weight = 2

    # counting the occurrences of the input notes
    counter = Counter(notes_std)

    # bulting the multidimensional affinity list
    affinities = []
    for i in range(len(mode_signatures)):  # for each mode signature (interval structure)
        affinities.append([])
        for j in range(7):  # TODO: adapt to scale length?
            curr_mode = modes_dict[root][i][j]  # list of 7 notes
            aff = 0  # initializing the affinity
            for k in range(7):
                # calculating the positive weights
                aff = aff + counter[curr_mode[k]] * positive_weights[k]
            not_in_the_mode = [note for note in notes_std if note not in curr_mode]
            aff = aff - len(not_in_the_mode) * negative_weight
            aff = aff / len(notes_std)  # normalizing
            affinities[i].append(aff)

    return affinities



from melodically.harmony import musical_notes, std_to_midi, midi_to_std, mode_signatures
import numpy as np

"""
Dictionary containing, for each chord, 
another dictionary that indicates 
the color tones (c) and the chord tones (l).
"""
chord_tones = {}


def get_chord_tones(chord_dict):
    """
    Adds all the major (XM) / minor (Xm) / dominant (X7)
    chords to the chord dictionary.

    :param chord_dict: chord dictionary
    :return: the updated chord dictionary
    """
    for root in musical_notes:
        # 0 -> XM (major) ; 4 -> X7 (dominant); 5 -> Xm (minor)
        for shift in [0, 4, 5]:
            # root in std notation
            root_midi = std_to_midi(root)

            # getting the indices of chord (c) and color (l) tones
            chord_tones_indices = [0, 2, 4]  # I III V
            color_tones_indices = [i for i in range(7) if i not in chord_tones_indices]

            # applying a cumulative sum to the diatonic scale intervals
            # circular shift used to switch from major, minor and dominant (np.roll)
            # [2, 2, 1, 2, 2, 2, 1] => [0, 2, 4, 5, 7, 9, 11]
            m0 = list(np.roll(mode_signatures[0], -1 * shift))
            diatonic_scale_absolute_intervals = [sum(m0[0:i]) for i, value in enumerate(m0)]

            # getting two lists of notes in std notation for c and l
            c_tones = [midi_to_std(root_midi + diatonic_scale_absolute_intervals[i]) for i in chord_tones_indices]
            l_tones = [midi_to_std(root_midi + diatonic_scale_absolute_intervals[i]) for i in color_tones_indices]

            # dominant chords have the VII in the chord tones
            if shift == 4:
                c_tones.append(l_tones.pop(3))

            # choosing the suffix
            if shift == 0:
                suffix = 'M'
            elif shift == 4:
                suffix = '7'
            else:
                suffix = 'm'

            chord_dict[root + suffix] = {
                'c': c_tones,
                'l': l_tones
            }

    return chord_dict


# adding major, minor and dominant chords to the dictionary
# TODO: check
chord_tones = get_chord_tones(chord_tones)


def chord_to_midi(chord, octave=3):
    # TODO: README update
    """
    Given a certain chord, it returns a list with the midi values
    of its notes. An octave parameter can be specified as a integer
    indicating the octave shared between the notes, as a list of integers,
    indicating the octave for each note of the chords (in this case, the size
    of the list must be the same of the number of chord tones of the chord).

    :param chord: target chord
    :param octave: octave of all the notes or of the single notes in case a list is passed
    :return: a list of midi note values of the chord tones of the chord
    """
    notes = chord_tones[chord]['c']
    if isinstance(octave, list):  # if octave is a list contains multiple values for each note
        midi_notes = [std_to_midi(n, o) for n, o in zip(notes, octave)]
    else:
        midi_notes = [std_to_midi(n, octave) for n in notes]
    return midi_notes



from melodically.midi_note_queue import *
from melodically.harmonic_state import *
from melodically.harmony import *
from melodically.chords import chord_tones, chord_to_midi
from melodically.rhythm import *
from melodically.parsers import *



def get_durations(bpm):
    """
    Function that generate a dictionary containing
    the duration in seconds of various rhythmic figures.

    :param bpm: beat per minutes
    :return: rhythmic dictionary
    """
    beat_time = 60 / bpm
    return {
        '1': beat_time * 4,
        '2': beat_time * 2,
        '4': beat_time * 1,
        '4dot': beat_time * 3 / 2,
        '4t': beat_time * 2 / 3,
        '8': beat_time * 1 / 2,
        '8t': beat_time * 1 / 3,
        '16': beat_time * 1 / 4,
        '16t': beat_time * 1 / 6,
    }


def get_nearest_rhythm(interval, rhythmical_durations):
    """
    Given a certain interval in seconds, gets the rhythmical duration
    that has the lower distance with it.

    :param interval: duration in seconds
    :param rhythmical_durations: dictionary returned by the get_durations
    :return:
    """

    # the keys and the values obtained from the methods "values" and "keys"
    # of a dictionary are correctly ordered, so we can calculate
    # the distances from the keys, get the index of the argmin
    # and use it to get the correct key

    # values of the rhythmical_durations dictionary
    rhythmical_durations_values = list(rhythmical_durations.values())

    # keys of the rhythmical_durations dictionary
    rhythmical_durations_keys = list(rhythmical_durations.keys())

    # list comprehension used to map the distance function to the values
    distances = [abs(interval - x) for x in rhythmical_durations_values]

    # argmin of the distance (an index)
    result_index = distances.index(min(distances))

    # using the index to get the correct rhythmical duration key
    return rhythmical_durations_keys[result_index]






from melodically.harmony import midi_to_std, get_root, harmonic_affinities, modes_dict


class HarmonicState:
    """
    This class allows to monitor and update an internal harmonic state,
    influenced by external notes inputs
    """

    def __init__(self, buffer_size=16):
        # contains the input notes in std notation
        self.noteBuffer = []

        # max size of the buffer
        self.bufferSize = buffer_size

        # mode dictionary used to represent a modal scale
        # root: root note of the scale
        # mode_signature_index: index of the interval scheme of a modal scale
        # mode_index: index of the mode inside a certain interval scheme
        self.currentMode = {'root': 'C', 'mode_signature_index': 0, 'mode_index': 0}

    def push_notes(self, new_notes):
        """
        Pushes new note inside the buffer.
        If the buffer overflows, the older notes are discarded.

        :param new_notes: list of new notes
        """
        self.noteBuffer = self.noteBuffer + new_notes
        while len(self.noteBuffer) > self.bufferSize:
            self.noteBuffer.pop(0)  # removing old notes

    def update_scale(self):
        """
        Updates the currentMode attribute based on the notes in the buffer,
        applying the harmonic_affinities function to them.

        :return: currentMode
        """
        notes_std = [n for n in self.noteBuffer]
        root = get_root(notes_std)
        modes_affinities = harmonic_affinities(root, notes_std)
        mode_signature_index = modes_affinities.index(max(modes_affinities))
        tmp = modes_affinities[mode_signature_index]  # sequence with the lowest distance
        mode_index = tmp.index(max(tmp))

        self.currentMode['root'] = root
        self.currentMode['mode_signature_index'] = mode_signature_index
        self.currentMode['mode_index'] = mode_index

        return self.currentMode

    def get_mode_notes(self):
        """
        Gets the current notes of the current modal scale.

        :return: list of notes of the current modal scale
        """
        self.update_scale()
        tmp = modes_dict[self.currentMode['root']]
        tmp = tmp[self.currentMode['mode_signature_index']]
        return tmp[self.currentMode['mode_index']]
